# Duo, A crazy eights game.
# Copyright (C) 2009 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""The low level game messaging service.

The transport layer implements a network gathering all the game
entities (players, referee, observers, ...). It is responsible to
forward messages from one entity to another one, or from one entity to
all the others. It doesn't assume anything about the embedded message.

The transport layer is interchangeable as long as it provides the
following services.

Transport quality
-----------------
The transport layer should be reliable and ordered. In other words, it
should check that all the messages reach the endpoint(s) to which they
were destined. Additionally, it should forward the messages in the same
order it has received them.

Endpoints and Authentication
----------------------------
All game entities can connect onto the transport layer using an
endpoint. Endpoints are designated by an ID. The ID of the endpoint
is provided by the upper (protocol) layer. This ID shall be unique
among all the connected endpoints.

Context Management
------------------
The transport layer should provide the notion of contexts. A context
is an environment which constrains the communication to the endpoints
that belong to it. An endpoint can belong to only one context at the
same time.

"""

__docformat__ = 'restructuredtext'

import cPickle
import logging
import os
import sys
import threading

from duo import tools


# The module must be imported after the logger was configured.
logger = logging.getLogger('duo.transport')


class EndpointAlreadyConnectedError(Exception):
    """Cconnection was refused because the client name is already in use."""
    pass


class UnknownEndpointError(Exception):
    """The client is not present on the transport layer."""
    pass


class BadEndpointError(Exception):
    """The endpoint is invalid, usually because it is null or of wrong type."""
    pass


class BadContextError(Exception):
    """The context is invalid, usually because it is null or of wrong type."""
    pass


class Packet(object):
    """The message wrapper at the transport level.

    A packet is defined by its type and its payload. Typing a packet
    is done by subclassing the Packet class. The payload depends on
    the kind of packet.
    """
    def __init__(self, payload):
        self._payload = payload

    def __str__(self):
        return '[' + str(self.__class__.__name__) + '] ' + str(self._payload)

    def _get_payload(self):
        return self._payload

    payload = property(_get_payload, None, None, 'The content of the packet')


class ProtocolPacket(Packet):
    """Protocol related packet.

    The PROTOCOL packet is a packet generated by the upper layer. Its
    payload is some protocol object.
    """
    def __init__(self, payload):
        Packet.__init__(self, payload)


class TransportJoinPacket(Packet):
    """Transport layer related packet.

    The TRANSPORT Join packet is generated by the transport layer when
    an endpoint joins a context. Its payload is the endpoint concerned
    with the event.
    """
    def __init__(self, payload):
        Packet.__init__(self, payload)


class TransportEscapePacket(Packet):
    """Transport layer related packet.

    The TRANSPORT Escape packet is generated by the transport layer
    when an endpoint is detached from a context. Its payload is the
    endpoint concerned with the event.
    """
    def __init__(self, payload):
        Packet.__init__(self, payload)


class Context(object):
    """An environment which isolates some endpoints from all the others.

    Contexts are stored as a hierarchy. When an endpoint creates a
    context, the context becomes the child of the current one. A
    context is automatically destroyed when the last endpoint
    connected to it leaves. The layer provides a default root context
    which cannot be destroyed.

    Endpoints always belong to exactly one context. Any endpoint can
    consult the list of existing context, and join one. Any endpoint can
    create a new context. Doing this, it is automatically moved into this
    new context.
    """

    def __init__(self, name, parent=None):
        self._parent = parent
        self._lock = threading.Lock()
        self._name = name
        self._endpoints = {}
        self._contexts = []

    def __str__(self):
        return 'C[' + self._name + ']'

    def add_endpoint(self, endpoint):
        """Add an endpoint to this context.

        Since the endpoint also has a reference to its current
        context, this method also updates the given endpoint.
        """
        self.add_endpoint_without_check(endpoint)
        try:
            endpoint.set_context_without_check(self)
        except AttributeError:
            raise BadEndpointError, '%s' % endpoint

    def add_endpoint_without_check(self, endpoint):
        """Add an endpoint to this context.

        This method leaves the endpoint untouched. It should only be
        used by an endpoint instance to prevent loops.
        """
        self._lock.acquire()
        try:
            if self._endpoints.has_key(endpoint):
                raise EndpointAlreadyConnectedError, '%s' % endpoint
            self._endpoints[endpoint.get_id()] = endpoint
            logger.debug('Inserted %s in %s' % (endpoint, self))
        finally:
            self._lock.release()

    def remove_endpoint(self, endpoint):
        """Remove an endpoint from this context.

        Since the endpoint also has a reference to its current
        context, this method also updates the given endpoint.
        """
        self.remove_endpoint_without_check(endpoint)
        try:
            if self._parent:
                self._parent.add_endpoint(endpoint)
        except AttributeError:
            raise BadEndpointError, '%s' % endpoint

    def remove_endpoint_without_check(self, endpoint):
        """Remove an endpoint from this context.

        This method leaves the endpoint untouched. It should only be
        used by an endpoint instance to prevent loops.
        """
        self._lock.acquire()
        try:
            try:
                self._endpoints.pop(endpoint.get_id())
                logger.debug('Removed %s from %s' % (endpoint, self))
            except KeyError:
                raise UnknownEndpointError()
        finally:
            self._lock.release()

    def get_endpoints(self):
        """Return the list of endpoints connected to this context.
        """
        self._lock.acquire()
        try:
            return self._endpoints.values()
        finally:
            self._lock.release()

    def get_parent(self):
        """Return the parent of this context.
        """
        self._lock.acquire()
        try:
            return self._parent
        finally:
            self._lock.release()

    def get_name(self):
        """Return the name of this context.
        """
        self._lock.acquire()
        try:
            return self._name
        finally:
            self._lock.release()

    def get_child_contexts(self):
        """Get the list of children to this context.
        """
        return self._contexts

    def get_or_create_context(self, name):
        """Retrieve the context whose name is given.
        """
        self._lock.acquire()
        try:
            if self._name == name:
                logger.debug('Found context %s' % name)
                return self

            for context in self.get_child_contexts():
                if context.get_name() == name:
                    logger.debug('Found context %s' % context)
                    return context

            context = Context(name, self)
            self._add_child_context(context)
            return context
        finally:
            self._lock.release()

    def _add_child_context(self, context):
        """TODO"""
        logger.debug('Added context %s' % context)
        self._contexts.append(context)

    def _remove_child_context(self, context):
        """TODO"""
        if context is None:
            raise BadContextError()
        elif context not in self._contexts:
            raise BadContextError()
        else:
            self._contexts.remove(context)


class Endpoint(object):
    """The proxy to any entity connected to the layer.

    Endpoints are an abstraction of persons, hosts, programs, anything
    which can be connected onto the layer. Endpoint can belong to only
    one context at any time. If the Endpoint doesn't belong to any
    context, it is said disconnected from the layer.

    TODO Describe callbacks dict.
    """
    def __init__(self, id, callbacks):
        self._context = None
        self._id = id
        self._cb = callbacks

    def __str__(self):
        return 'E[' + self._id + ']'

    def get_id(self):
        """Return the unique ID of this endpoint."""
        return self._id

    def set_context(self, context):
        if self._context:
            self._context.remove_endpoint_without_check(self)
        self.set_context_without_check(context)

    def set_context_without_check(self, context):
        logger.debug('Attached %s to %s' % (self, context))
        self._context = context

    def get_context(self):
        return self._context

    def notify(self, byte_array, emitter):
        """Notify the reception of a message."""
        packet = cPickle.loads(byte_array)
        self._do_notify(packet.__class__, packet.payload, emitter)

    def send_to(self, payload, to):
        """Send a message to another endpoint."""
        self._do_send(cPickle.dumps(ProtocolPacket(payload)), to)

    def broadcast(self, payload):
        """Send a message to every connected endpoints in the current context.
        """
        self._do_send(cPickle.dumps(ProtocolPacket(payload)), None)

    def get_contexts(self):
        """Return full information on existing contexts."""
        raise NotImplementedError, \
              '%s shall define notify' % self.__class__.__name__

    def join_context(self, name):
        """Attach this endpoint to the given context."""
        raise NotImplementedError, \
              '%s shall define notify' % self.__class__.__name__

    def escape(self):
        """Detach from the current context.

        Once detached, the endpoint is automatically attached to the
        parent context. If there is no parent, the endpoint is disconnected.
        """
        raise NotImplementedError, \
              '%s shall define notify' % self.__class__.__name__

    def _do_send(self, packet, to):
        """Send a packet.

        If aTo is None, the packet has to be broadcasted.
        """
        raise NotImplementedError, \
              '%s shall define notify' % self.__class__.__name__

    def _do_notify(self, cb_key, payload, emitter):
        try:
            cb = self._cb[cb_key]
            #logger.debug('sending notification to %s on %s '
            #             % (self.get_id(), cb.__name__))
            cb.__call__(payload, emitter)
        except KeyError:
            pass
        except Exception, e:
            logger.error(e)


def get_layer_options(layer):
    """Return the available options for a transport layer.

    Options are specific to the transport layers. They are facultative and
    are useful to provide configuration parameters to the selected layer.
    """
    try:
        module = tools.import_recursively('duo.transport.' + layer)
        return module.OPTIONS
    except Exception, e:
        logger.error('getLayerOptions %s: %s' % (layer, e))


def create_endpoint(layer, layer_options, id):
    try:
        module = tools.import_recursively('duo.transport.' + layer)
        return module.CLASS(id, layer_options)
    except Exception, e:
        logger.error('createEndpoint %s: %s' % ((layer, layer_options), e))


# Test all the available layers
def test():

    import time

    test_logger = logging.getLogger('duo.test')

    def on_message(message, emitter):
        test_logger.info('Received \'%s\' from %s' % (message, emitter))

    def on_connection(peer, emitter):
        test_logger.info('%s is now connected' % peer)

    def on_disconnection(peer, emitter):
        test_logger.info('%s is now disconnected' % peer)

    callbacks = {TransportJoinPacket: on_connection,
                 TransportEscapePacket: on_disconnection,
                 ProtocolPacket: on_message }

    layers = get_layers()
    for layer in layers:

        time.sleep(1)
        test_logger.info('=== Testing %s ===' % layer)
        # TODO: fill in the layer options
        options = { 'context': 'root',
                    'callbacks': callbacks }

        test_logger.info('--- Creating endpoints ---')
        a = create_endpoint(layer, options, 'A')
        b = create_endpoint(layer, options, 'B')
        c = create_endpoint(layer, options, 'C')

        time.sleep(1)
        test_logger.info('--- Testing transport ---')
        a.send_to('hello B', 'B')
        b.broadcast('hello all')

        time.sleep(1)
        test_logger.info('--- Disconnecting endpoints ---')
        a.escape()
        b.escape()
        c.escape()
